/**
 * An ADC client emulator.
 */

#include "uhub.h"

#define ADC_CLIENTS_DEFAULT 2
#define ADC_CLIENTS 750

#define ADC_BUFSIZE 16384
#define ADC_SIDSIZE 4
#define ADC_CID_SIZE 39

#define BIG_BUFSIZE 131072
#define TIGERSIZE 24

#define ADC_HANDSHAKE "HSUP ADBASE ADTIGR ADPING\n"
#define ADCRUSH "adcrush/0.1"

static struct sockaddr_in saddr;
static int debug = 1;

static void create_sockaddr(const char* host, uint16_t port)
{
	memset(&saddr, 0, sizeof(saddr));
	inet_aton(host, &saddr.sin_addr);
	saddr.sin_family = AF_INET;
	saddr.sin_port   = htons(port);
}

struct ADC_client
{
	int  sd;
	sid_t sid;
	enum user_state state;
	char info[ADC_BUFSIZE];
	char recvbuf[BIG_BUFSIZE];
	char sendbuf[BIG_BUFSIZE];
	size_t s_offset;
	size_t r_offset;
	struct event ev_read;
	struct event ev_write;
	struct event ev_timer;
	size_t timeout;
};

static void adc_cid_pid(struct ADC_client* client)
{
	char seed[64];
	char pid[64];
	char cid[64];
	uint64_t tiger_res1[3];
	uint64_t tiger_res2[3];

	/* create cid+pid pair */
	memset(seed, 0, 64);
	snprintf(seed, 64, ADCRUSH "%p/%d", client, (int) time(0));
	
	tiger((uint64_t*) seed, strlen(seed), tiger_res1);
	base32_encode((unsigned char*) tiger_res1, TIGERSIZE, pid);
	tiger((uint64_t*) tiger_res1, TIGERSIZE, tiger_res2);
	base32_encode((unsigned char*) tiger_res2, TIGERSIZE, cid);
	
	cid[ADC_CID_SIZE] = 0;
	pid[ADC_CID_SIZE] = 0;
	
	strcat(client->info, " PD");
	strcat(client->info, pid);
	strcat(client->info, " ID");
	strcat(client->info, cid);
}

static size_t get_wait_rand(size_t max)
{
	static size_t next = 0;
	if (next == 0) next = time(0);
	next = next * 1103515245 + 12345;
	return((size_t )(next / 65536) % max);
}

static void set_state_timeout(struct ADC_client* client, enum user_state state)
{
	size_t next_timeout = 0;
	
	switch (state)
	{
		case state_protocol: next_timeout = 30; break;
		case state_identify: next_timeout = 30; break;
		case state_verify:   next_timeout = 30; break;
		case state_normal:   next_timeout = 120; break;
		case state_cleanup:  next_timeout = 120; break;
		case state_disconnected: next_timeout = 120; break;
	}
	
	client->state   = state;
	client->timeout = get_wait_rand(next_timeout);
};

static void send_client(struct ADC_client* client, char* msg)
{
	int ret = net_send(client->sd, msg, strlen(msg), 0);
	
	if (debug)
	{
		char* dump = strdup(msg);
		dump[strlen(msg) - 1] = 0;
		printf("SEND: '%s'\n", dump);
		free(dump);
	}
	
 	if (ret != strlen(msg))
	{
		printf("ret (%d) != msg->length (%d)\n", ret, (int) strlen(msg));
	}
}


static void send_client_handshake(struct ADC_client* client)
{
	send_client(client, ADC_HANDSHAKE);
	set_state_timeout(client, state_protocol);
}

static void send_client_info(struct ADC_client* client)
{
	client->info[0] = 0;
	strcat(client->info, "BINF ");
	strcat(client->info, sid_to_string(client->sid));
	strcat(client->info, " NIadcrush_");
	strcat(client->info, sid_to_string(client->sid));
	strcat(client->info, " VE" ADCRUSH);
	strcat(client->info, " DEcrash\\stest\\sdummy");
	strcat(client->info, " I40.0.0.0");
	strcat(client->info, " EMuhub@extatic.org");
	strcat(client->info, " SL3");
	strcat(client->info, " HN1");
	strcat(client->info, " HR1");
	strcat(client->info, " HO1");
	
	adc_cid_pid(client);
	
	strcat(client->info, "\n");
	
	send_client(client, client->info);
}

static int recv_client(struct ADC_client* client)
{
	ssize_t size = net_recv(client->sd, &client->recvbuf[client->r_offset], ADC_BUFSIZE - client->r_offset, 0);
	if (size == 0 || ((size == -1 && errno != EAGAIN)))
		return -1;
	client->recvbuf[client->r_offset + size] = 0;

	char* start = client->recvbuf;
	char* pos;
	char* lastPos;
	while ((pos = strchr(start, '\n')))
	{
		lastPos = pos;
		pos[0] = 0;
		
		if (debug)
		{
			printf("RECV: '%s'\n", start);
		}
		
		fourcc_t cmd = 0;
		if (strlen(start) < 4)
		{
			puts("crap from server");
			start = &pos[1];
			continue;
		}
		
		cmd = FOURCC(start[0], start[1], start[2], start[3]);
		
		switch (cmd)
		{
			case ADC_CMD_ISUP:
				break;

			case ADC_CMD_ISID:
				if (client->state == state_protocol)
				{
					client->sid = string_to_sid(&start[5]);
					client->state = state_identify;
					
				}
				break;
				
			case ADC_CMD_IINF:
				if (client->state == state_identify)
				{
					client->state = state_identify;
					send_client_info(client);
				}
				break;
				
			case ADC_CMD_ISTA:
				break;
				
			default:
				/* printf("????: '%s'\n", start); */
				break;
		}
		
		start = &pos[1];
	}
	
	client->r_offset = strlen(lastPos);
	memmove(client->recvbuf, lastPos, strlen(lastPos));
	memset(&client->recvbuf[client->r_offset], 0, ADC_BUFSIZE-client->r_offset);
	
	
	return 0;
}

void ADC_client_connect(struct ADC_client* client)
{
	struct timeval timeout = { TIMEOUT_IDLE, 0 };
	
	net_connect(client->sd, (struct sockaddr*) &saddr, sizeof(struct sockaddr_in));
	
	event_add(&client->ev_read, &timeout);
	event_add(&client->ev_write, &timeout);
}


void event_callback(int fd, short ev, void *arg)
{
	struct ADC_client* client = (struct ADC_client*) arg;

	if (ev & EV_WRITE)
	{
		switch (client->state)
		{
			case state_normal:       break;
			case state_disconnected: send_client_handshake(client); break;
			default:
				break;
		}
	}
	
	if (ev & EV_TIMEOUT)
	{
		if (client->state == state_disconnected)
		{
			ADC_client_connect(client);
		}
	}
	
	if (ev & EV_READ)
	{
		if (recv_client(client) == -1)
		{
			event_del(&client->ev_read);
		}
	}
}

int ADC_client_create(struct ADC_client* client)
{
	struct timeval timeout = { 0, 0 };
	
	memset(client, 0, sizeof(struct ADC_client));
	client->sd = net_socket_create(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (client->sd == -1) return -1;
	
	set_state_timeout(client, state_disconnected);
	
	event_set(&client->ev_write, client->sd, EV_WRITE, event_callback, client);
	event_set(&client->ev_read,  client->sd, EV_READ | EV_PERSIST, event_callback, client);
	
	net_set_nonblocking(client->sd, 1);
	
	timeout.tv_sec = client->timeout;
	evtimer_set(&client->ev_timer, event_callback, client);
	evtimer_add(&client->ev_timer, &timeout);

	return 0;
}

int ADC_client_destroy(struct ADC_client* client)
{
	if (client->sd != -1)
	{
		net_close(client->sd);
		client->sd = -1;
		return 0;
	}
	return -1;
}


void runloop(size_t clients)
{
	size_t n = 0;
	struct ADC_client* client[ADC_CLIENTS];

	for (n = 0; n < clients; n++)
	{
		struct ADC_client* c = malloc(sizeof(struct ADC_client));
		client[n] = c;

		ADC_client_create(c);
		if (n == 0)
		{
			ADC_client_connect(c);
			evtimer_del(&c->ev_timer);
		}
	}
	
	event_dispatch();
	
	for (n = 0; n < clients; n++)
	{
		ADC_client_destroy(client[n]);
		free(client[n]);
	}
}


static void print_usage(const char* program)
{
	printf("Usage: %s host port [connections]\n", program);
}


int main(int argc, char** argv)
{
	int clients = ADC_CLIENTS_DEFAULT;
	struct hostent* dns;
	struct in_addr* addr;
	char* host;
	int port;

	if (argc < 3)
	{
		print_usage(argv[0]);
		return -1;
	}

	port = strtol(argv[2], NULL, 10);
	dns = gethostbyname(argv[1]);
	if (dns)
	{
		addr = (struct in_addr*) dns->h_addr_list[0];
		host = inet_ntoa(*addr);
	}
	else
	{
		host = NULL;
	}
	
	if (!host || port < 1 || port > 65535)
	{
		print_usage(argv[0]);
		return -2;
	}

	if (argc > 3)
	{
		clients = strtol(argv[3], NULL, 10);
		if (clients > ADC_CLIENTS || clients == 0) clients = ADC_CLIENTS_DEFAULT;
	}
	
	net_initialize();
	
	create_sockaddr(host, port);
	
	runloop(clients);

	net_shutdown();

	return 0;
}

